"""
Data Manager
============

The DataManager is a wrapper around all data sources used by RUINSapp.
It can be configures by any :class:`Config <ruins.core.config.Config>` class
and organizes or caches all data sources using a 
:class:`DataSource <ruins.core.data_manager.DataSource>` inherited class.
This makes the read and filter interface available on all sources, no matter
where they are stored.
Using the :class:`Config <ruins.core.config.Config>` to instantiate a data
manager can in principle enabled different profiles, or even an interaction
with the frontend, although not implemented nor desired at the current stage.

Example
-------

.. code-block:: python

    from ruins import core

    # create default config
    conf = core.Config()

    # create a data manager from this
    dm = core.DataManager(**conf)

Of course, the data manager can also be used without the config, ie. to open it
in debug mode:

.. code-block:: python
    
    # using conf with conf.debug=False and overwrite it
    dm = core.DataManager(**conf, debug=True)

"""
import abc
import os
import glob
import inspect
import xarray as xr
import pandas as pd
from collections.abc import Mapping
from typing import Type, List



class DataSource(abc.ABC):
    """
    Abstract base class for data sources. This provides the common interface
    for data sources of different source types (like file, URL, database).
    """
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    @abc.abstractmethod
    def read(self):
        pass
    
    @abc.abstractmethod
    def filter(self, **kwargs):
        pass


class FileSource(DataSource, abc.ABC):
    """
    Abstract base class for file sources. This provides the common interface
    for every data source that is based on a file.
    """
    def __init__(self, path: str, cache: bool = True, hot_load = False, **kwargs):
        super().__init__(**kwargs)
        self.path = path
        self.cache = cache
        
        # check if the dataset should be pre-loaded
        if hot_load:
            self.cache = True
            self.data = self._load_source()

    @abc.abstractmethod
    def _load_source(self):
        """Method to load the actual source on the disk"""
        pass

    def read(self):
        if self.cache:
            if not hasattr(self, 'data'):
                self.data = self._load_source()
            return self.data

        else:
            return self._load_source()
    
    def filter(self):
        pass


class HDF5Source(FileSource):
    """
    HDF5 file sources. This class is used to load HDF5 files.
    """
    def _load_source(self) -> xr.Dataset:
        return xr.open_dataset(self.path)
    
    def read(self) -> xr.Dataset:
        return super(HDF5Source, self).read()


class CSVSource(FileSource):
    """
    CSV file source. This class is used to load CSV files.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # inspect read_csv to learn about allowed param
        sig = inspect.signature(pd.read_csv)
        self.pandas_params = list(sig.parameters.keys())

    def _load_source(self):
        # extract pandas args
        pandas_args = {k: v for k, v in self._kwargs.items() if k in self.pandas_params}

        # load data
        return pd.read_csv(self.path, **pandas_args)


class DataManager(Mapping):
    """Main class for accessing different data sources.

    The DataManager holds and manages all data sources. The default behavior is
    to scan the specified path for files of known file extension and cache them
    in memory.

    Parameters
    ----------
    datapath : str
        A location where the data is stored. The class will load all sources 
        there and make them accessible through DataSource classes.
    cache : bool
        Will be passed to the DataSource classes. It true, the source will only
        be read once and then stored in memory until the DataManager gets
        deconstructed.
    include_mimes : dict
        A dictionary of file extensions and their corresponding DataSource.
        If something is not listed, the DataManager will ignore the file type.
        The include_mimes can be overwritten by passing filenames directly.

    """
    def __init__(self, datapath: str = None, cache: bool = True, hot_load = False, debug: bool = False, **kwargs) -> None:
        """
        You can pass in a Config as kwargs.
        """
        # check if no config - or config without datapath - was passed
        if datapath is None:
            from ruins.core import Config
            self.from_config(**Config(**kwargs))
        else:
            self.from_config(datapath=datapath, cache=cache, hot_load=hot_load, debug=debug, **kwargs)
    
    def read(self, name_or_file: str):
        # if config init then there is an attrribute called datafile_names
        if 'datafile_names' in self._config:
            # if name_or_file exists in Config.datafile_names, the filename stored there is used. Otherwise name_or_file is considered the filename.
            identifier = self._config['datafile_names'].get(name_or_file, name_or_file) # unittest
        else:
            identifier = name_or_file

        return self[identifier].read()

    def from_config(self, datapath: str = None, cache: bool = True, hot_load: bool = False, debug: bool = False, **kwargs) -> None:
        """
        Initialize the DataManager from a :class:`Config <ruins.core.Config>` object.
        """
        # store the main settings
        self._config = kwargs
        self._datapath = datapath
        self.cache = cache
        self.hot_load = hot_load
        self.debug = debug

        # file settings
        self._data_sources = {}

        # infer data source
        if self._datapath is not None:
            self._infer_from_folder()
    
    @property
    def datapath(self) -> str:
        return self._datapath

    @datapath.setter
    def datapath(self, path: str) -> None:
        if os.path.exists(path):
            self._datapath = path
            self._infer_from_folder()
        else:
            raise OSError(f"{path} does not exist.")
    
    @property
    def datasources(self) -> List[DataSource]:
        return list(self._data_sources.keys())

    def _infer_from_folder(self) -> None:
        """
        Read all files from the datapath as specified on instantiation.
        Calls :func:`add_source` on each file.
        """
        # get a list of all files
        file_list = glob.glob(os.path.join(self.datapath, '*'))
        file_list.extend(glob.glob(os.path.join(self.datapath, '**', '*')))


        for fname in file_list:
            self.add_source(path=fname, not_exists='warn' if self.debug else 'ignore')

    def add_source(self, path: str, not_exists: str = 'raise') -> None:
        """
        Add a file as data source to the DataManager.
        Only if the file has an allowed file extension, it will be managed.
        Files of same name will be overwritten, this is also true if they had
        different extensions.

        """
        # load the tracked source base class
        mimes = self._config.get('default_sources', {})

        # check if the config holds arguments for this source instance
        args = self._config.get('sources_args', {}).get(os.path.basename(path), {})

        # get the basename
        try:
            basename, mime = os.path.basename(path).split('.')
        except ValueError:
            if self.debug:
                print(f"[Warning]: {path} has no extension.")
            return 
        
        if mime in mimes.keys():
            # get the class - overwirte by direct kwargs settings if needed
            clsName = mimes[mime] if basename not in self._config else self._config[basename]
            BaseClass = self.resolve_class_name(clsName)
            
            # add the source
#            args = self._config.get(basename, {})
            args.update({'path': path, 'cache': self.cache, 'hot_load': self.hot_load})
            self._data_sources[basename] = BaseClass(**args)
        else:
            if not_exists == 'raise':
                raise OSError(f"{path} is not a configured data source")
            elif not_exists == 'ignore':
                pass
            elif not_exists == 'warn':
                print(f"{path} is found, but not a configured data source")

    def resolve_class_name(self, cls_name: str) -> Type[DataSource]:
        # checkout globals
        cls = globals().get(cls_name, False)
        
        # do we have a class?
        if not cls:
            # TODO, there is maybe an extension module to search one day
            raise RuntimeError(f"Can't find class {cls_name}.")
        
        return cls

    def __len__(self):
        """Return the number of managed data sources"""
        return len(self._data_sources)

    def __iter__(self):
        """Iterate over all dataset names"""
        for name in self._data_sources.keys():
            yield name
    
    def __getitem__(self, key: str) -> DataSource:
        """Return the requested datasource"""
        return self._data_sources[key]

    def __repr__(self):
        return f"{self.__class__.__name__}(datapath={self.datapath}, cache={self.cache})"
    
    def __str__(self):
        return f"<DataManager of {len(self)} sources>"



from typing import Callable, Dict, Union
import os
from os.path import join as pjoin
import json
from collections.abc import Mapping

from ruins.core.i18n import get_translator

from streamlit import session_state
import streamlit as st

# check if streamlit is running
if not st._is_running_with_streamlit:
    session_state = dict()

class Config(Mapping):
    """
    Streamlit app Config object.

    This class holds all configs needed to run the apps. 
    It can be instantiated just like this to load default
    values. 
    If a path is provided, it will load the configs from
    the referenced json file. Any config can be updated
    by passed kwargs.

    This design makes the config updateable and easy to
    to manage. At the same time it can be persisted to
    the disk and even mirrored to a database if needed in
    the future.

    """
    def __init__(self, path: str = None, **kwargs) -> None:
        # set the default values

        # debug mode
        self._debug = False
        self.lang = 'en'

        # path 
        self.basepath = os.path.abspath(pjoin(os.path.dirname(__file__), '..', '..'))
        self.datapath = pjoin(self.basepath, 'data')
        self.hot_load = kwargs.get('hot_load', False)

        # datafile names, without file extension
        self.datafile_names = {
            'stations': 'stats',
            'cordex_grid': 'CORDEXgrid',
            'cimp_grid': 'CIMP5grid',
            'weather': 'weather',
            'climate': 'cordex_krummh'
            #'climate_coast': 'cordex_coast',
            #'hydro': 'hydro_krummh'
        }

        # mime readers
        self.default_sources = {
            'nc': 'HDF5Source',
            'csv': 'CSVSource',
        }
        self.default_sources.update(kwargs.get('include_mimes', {}))

        # reader args
        self.sources_args = {
            'stats.csv': dict(index_col=0),
            'hsim_collect.csv': dict(index_col=0),
            'windpowerx.csv': dict(index_col=0),
            'estQ.csv': dict(index_col=[0], parse_dates=[0]),
            'levelknock.csv': dict(index_col=[0], parse_dates=[0]),
            'levelW.csv': dict(index_col=[0], parse_dates=[0]),
            'prec.csv': dict(index_col=[0], parse_dates=[0]),
            'Qknock.csv': dict(index_col=[0], parse_dates=[0]),
        }
        self.sources_args.update(kwargs.get('include_args', {}))

        # app management
        self.layout = 'centered'

        # store the keys
        self._keys = ['debug', 'lang', 'basepath', 'datapath', 'hot_load', 'datafile_names', 'default_sources', 'sources_args', 'layout']

        # check if a path was provided
        conf_args = self.from_json(path) if path else {}

        # update with kwargs
        conf_args.update(kwargs)
        self._update(conf_args)

    @property
    def debug(self):
        return self._debug

    @property
    def story_mode(self):
        return self._story_mode
    
    @debug.setter
    def debug(self, value: Union[str, bool]):
        if isinstance(value, str):
            self._debug = value.lower() != 'false'
        else:
            self._debug = bool(value)

    @story_mode.setter
    def story_mode(self, value: Union[str, bool]):
        if isinstance(value, str):
            self._story_mode = value.lower() != 'false'
        else:
            self._story_mode = bool(value)

    def from_json(self, path: str) -> dict:
        """loads the content of the JSON config file"""
        if os.path.exists(path):
            with open(path, 'r') as f:
                return json.load(f)
        else:
            raise AttributeError(f"Config file {path} does not exist")
    
    def _update(self, new_settings: dict) -> None:
        """Update this instance with new settings"""
        for k, v in new_settings.items():
            setattr(self, k, v)
            if k not in self._keys:
                self._keys.append(k)
    
    def get_control_policy(self, control_name: str) -> str:
        """
        Get the control policy for the given control name.

        allowed policies are:
            - show: always show the control on the main container
            - hide: hide the control on the main container, but move to the expander
            - ignore: don't show anything

        """
        if self.has_key(f'{control_name}_policy'):
            return self.get(f'{control_name}_policy')
        elif self.has_key('controls_policy'):
            return self.get('controls_policy')
        else:
            # TODO: discuss with conrad to change this
            return 'show'

    def translator(self, **translations: Dict[str, str]) -> Callable[[str], str]:
        """Return a translator function"""
        return get_translator(self.lang, **translations)
    
    def get(self, key: str, default = None):
        if hasattr(self, key):
            return getattr(self, key)
        elif hasattr(session_state, key):
            return getattr(session_state, key)
        elif key in session_state:
            return session_state[key]
        else:
            return default
    
    def has_key(self, key) -> bool:
        if hasattr(self, key) and not key in session_state:
            session_state[key] = getattr(self, key)
        return hasattr(self, key) or hasattr(session_state, key) or key in session_state
    
    def __len__(self) -> int:
        return len(self._keys)

    def __iter__(self):
        for k in self._keys:
            yield k
    
    def __getitem__(self, key: str):
        if hasattr(self, key):
            return getattr(self, key)
        elif key in session_state:
            return session_state[key]
        else:
            raise KeyError(f"Key {key} not found")
    
    def __setitem__(self, key: str, value):
        setattr(self, key, value)
        if key not in self._keys:
            self._keys.append(key)



"""
Build a :class:`Config <ruins.core.Config>` and a 
:class:`DataManager <ruins.core.DataManager>` from a kwargs dict.
"""
from typing import Union, Tuple, Dict, List
import os
import shutil
import requests
import io
import zipfile
import streamlit as st

from .config import Config
from .data_manager import DataManager


st.experimental_singleton
def contextualized_data_manager(**kwargs) -> DataManager:
    return DataManager(**kwargs)


def build_config(omit_dataManager: bool = False, url_params: Dict[str, List[str]] = {}, **kwargs) -> Tuple[Config, Union[None, DataManager]]:
    """
    """
    # prepare the url params, if any
    # url params are always a list: https://docs.streamlit.io/library/api-reference/utilities/st.experimental_get_query_params
    # TODO: This should be sanitzed to avoid injection attacks!
    ukwargs = {k: v[0] if len(v) == 1 else v for k, v in url_params.items()}
    kwargs.update(ukwargs)

    # extract the DataManager, if it was already instantiated
    if 'dataManager' in kwargs:
        dataManager = kwargs.pop('dataManager')
    else:
        dataManager = None

    # build the Config
    config = Config(**kwargs)

    if omit_dataManager:
        return config,  None
    else:
        if dataManager is None:
            dataManager = contextualized_data_manager(**config)
        return config, dataManager


def download_data_archive(path: str = None, url: str = 'http://116.203.189.3/data.zip', DOI: str = None, if_exists: str = 'error'):
    """Download the data archive and extract into the data folder.
    If the path is None, the default path inside the repo itself is used.
    Then, you also need to change the datapath property of the application config.
    If the data folder already exists and is not empty, the function will error on default.
    You can pass ``if_exists='prune'`` to remove the existing data folder and replace it with the new one.
    """
    # use default path if none was provided
    if path is None:
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..', 'data'))
    
    # check if the data folder already exists
    if os.path.exists(path) and len(os.listdir(path)) > 0:
        if if_exists == 'error':
            raise OSError(f"The data path {path} already exists and is not empty. Pass if_exists='prune' to remove it.")
        elif if_exists == 'prune':
            shutil.rmtree(path)
            os.mkdir(path)
        else:
            raise AttributeError(f'if_exists must be one of "error", "prune"')
    
    # check which download route is used:
    if DOI is None:
        # now the data folder exists - download the archive
        print(f'Found Server URL: {url}\nStart downloading...', end='', flush=True)
        
        req = requests.get(url, stream=True)
        zip = zipfile.ZipFile(io.BytesIO(req.content))

        print(f'done.\nExtracting to {path}...', end='', flush=True)
        zip.extractall(os.path.abspath(os.path.join(path, '..')))
        print('done.', flush=True)
    else:
        # now the data folder exists - download the archive
        print(f'Found DOI: {DOI}\nStart downloading...', end='', flush=True)

        # Build the URL from Zenodo DOI
        chunk = DOI.split('/')[-1]
        record = chunk.split('.')[1]

        # request the existing data from Zenodo API
        dat = requests.get(f'https://zenodo.org/api/records/{record}').json()
        for f in dat['files']:
            if f['type'] == 'zip':
                req = requests.get(f['links']['self'], stream=True)
                zip = zipfile.ZipFile(io.BytesIO(req.content))

                # extract the data to the data folder
                print(f'done.\nExtracting to {path}...', end='', flush=True)
                zip.extractall(os.path.abspath(os.path.join(path, '..')))
                print('done.', flush=True)
                break
